==> OpenCL-0.8.4.tgz: OpenCL-0.8.4/src/nr.cpp <==
/*************************************************
* Nyberg-Rueppel Source File                     *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/nr.h>
#include <opencl/numthry.h>
#include <opencl/primes.h>

namespace OpenCL {

/*************************************************
* NR_PublicKey Constructor                       *
*************************************************/
NR_PublicKey::NR_PublicKey(const BigInt& prime, const BigInt& q1,
                           const BigInt& generator, const BigInt& y1) :
   p(prime), q(q1), g(generator), y(y1),
   powermod_g_p(g, p), powermod_y_p(y, p)
   {
   if(p <= 0 || q <= 0 || g <= 1)
      throw Invalid_Argument("NR_PublicKey: Invalid prime or generator");
   }

/*************************************************
* Check Public Nyberg-Rueppel Parameters         *
*************************************************/
bool NR_PublicKey::check_params() const
   {
   if((p - 1) % q != 0)
      return false;
   if(g <= 1 || g >= p)
      return false;
   if(!verify_prime(p) || !verify_prime(q))
      return false;
   return true;
   }

/*************************************************
* Nyberg-Rueppel Verification Function           *
*************************************************/
SecureVector<byte> NR_PublicKey::verify(const byte sig[], u32bit sig_len) const
   {
   if(sig_len != 2*q.bytes())
      return false;

   c.binary_decode(sig, sig_len / 2);
   d.binary_decode(sig + sig_len / 2, sig_len / 2);

   if(c.is_zero() || c >= q || d >= q)
      throw Invalid_Argument("Nyberg-Rueppel: Invalid signature");

   BigInt i = (powermod_g_p(d) * powermod_y_p(c)) % p;
   BigInt f = (c - i) % q;

   return encode(f);
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const BigInt& prime, const BigInt& q1,
                             const BigInt& generator, const BigInt& x1,
                             const BigInt& y1) :
   NR_PublicKey(prime, q1, generator, y1), x(x1)
   {
   if(!x && y != 0)
      throw Invalid_Argument("NR_PrivateKey: Public key cannot be set "
                             "without private key");

   if(!x)
      x = random_integer(1, q - 1);
   if(!y)
      {
      y = powermod_g_p(x);
      powermod_y_p = FixedBase_Exp(y, p);
      }
   }

/*************************************************
* Create a Nyberg-Rueppel Key                    *
*************************************************/
NR_PrivateKey::NR_PrivateKey(u32bit pbits)
   {
   if(pbits < 64)
      throw Invalid_Argument("NR_PrivateKey: Invalid sizes for primes");

   q = random_prime(2 * dl_work_factor(pbits));
   do
      p = random_integer(pbits - q.bits()) * q + 1;
   while(p.bits() != pbits || !is_prime(p));

   BigInt e = (p - 1) / q;
   u32bit prime_no = 0;
   do
      g = power_mod(PRIMES[prime_no++], e, p);
   while(g == 1);
   powermod_g_p = FixedBase_Exp(g, p);

   x = random_integer(q.bits() - 1);
   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, p);
   }

/*************************************************
* Check Private Nyberg-Rueppel Parameters        *
*************************************************/
bool NR_PrivateKey::check_params() const
   {
   if(!NR_PublicKey::check_params())
      return false;
   if(y != powermod_g_p(x))
      return false;
   return true;
   }

/*************************************************
* Nyberg-Rueppel Signature Operation             *
*************************************************/
SecureVector<byte> NR_PrivateKey::sign(const byte in[], u32bit len) const
   {
   BigInt f;
   f.binary_decode(in, len);

   if(f > q)
      throw Invalid_Argument("NR_PrivateKey::sign: Input is too large");

   while(true)
      {
      BigInt k(BigInt::Random, q.bits());
      while(k >= q)
         k.randomize(q.bits());
      c = (powermod_g_p(k) + f) % q;
      d = (k - x * c) % q;

      if(c.is_zero())
         continue;

      SecureVector<byte> output(2*q.bytes());
      c.binary_encode(output + (q.bytes() - c.bytes()));
      d.binary_encode(output + output.size() / 2 + (q.bytes() - d.bytes()));
      return output;
      }
   }

}

==> OpenCL-0.8.5.tgz: OpenCL-0.8.5/src/nr.cpp <==
/*************************************************
* Nyberg-Rueppel Source File                     *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/nr.h>
#include <opencl/numthry.h>
#include <opencl/primes.h>

namespace OpenCL {

/*************************************************
* NR_PublicKey Constructor                       *
*************************************************/
NR_PublicKey::NR_PublicKey(const DL_Group& domain, const BigInt& key) :
   params(domain), p(params.get_p()), q(params.get_q()), g(params.get_g()),
   y(key), powermod_g_p(g, p), powermod_y_p(y, p)
   {
   if(y < 0 || y >= p)
      throw Invalid_Argument("NR_PublicKey: Invalid public parameter");
   }

/*************************************************
* NR_PublicKey Copy Constructor                  *
*************************************************/
NR_PublicKey::NR_PublicKey(const NR_PublicKey& key) :
   PK_Verifying_with_MR_Key(),
   params(key.params), p(params.get_p()), q(params.get_q()), g(params.get_g()),
   y(key.get_y()), powermod_g_p(g, p), powermod_y_p(y, p)
   {
   }

/*************************************************
* Check Public Nyberg-Rueppel Parameters         *
*************************************************/
bool NR_PublicKey::check_params() const
   {
   if(y < 2 || y >= p)
      return false;
   if(!params.verify())
      return false;
   return true;
   }

/*************************************************
* Nyberg-Rueppel Verification Function           *
*************************************************/
SecureVector<byte> NR_PublicKey::verify(const byte sig[], u32bit sig_len) const
   {
   if(sig_len != 2*q.bytes())
      return false;

   c.binary_decode(sig, sig_len / 2);
   d.binary_decode(sig + sig_len / 2, sig_len / 2);

   if(c.is_zero() || c >= q || d >= q)
      throw Invalid_Argument("Nyberg-Rueppel: Invalid signature");

   BigInt i = (powermod_g_p(d) * powermod_y_p(c)) % p;
   BigInt f = (c - i) % q;

   return encode(f);
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& domain) : NR_PublicKey(domain, 0)
   {
   x = random_integer(1, q - 1);
   y = powermod_g_p(x);
   powermod_y_p = FixedBase_Exp(y, p);
   }

/*************************************************
* NR_PrivateKey Constructor                      *
*************************************************/
NR_PrivateKey::NR_PrivateKey(const DL_Group& domain,
                             const BigInt& priv, const BigInt& pub) :
   NR_PublicKey(domain, pub), x(priv)
   {
   if(x <= 1 || y <= 1 || x >= p || y >= p)
      throw Invalid_Argument("NR_PrivateKey: invalid private key");
   }

/*************************************************
* Check Private Nyberg-Rueppel Parameters        *
*************************************************/
bool NR_PrivateKey::check_params() const
   {
   if(!NR_PublicKey::check_params())
      return false;
   if(y != powermod_g_p(x))
      return false;
   return true;
   }

/*************************************************
* Nyberg-Rueppel Signature Operation             *
*************************************************/
SecureVector<byte> NR_PrivateKey::sign(const byte in[], u32bit len) const
   {
   BigInt f;
   f.binary_decode(in, len);

   if(f > q)
      throw Invalid_Argument("NR_PrivateKey::sign: Input is too large");

   while(true)
      {
      BigInt k(Random, q.bits());
      while(k >= q)
         k.randomize(q.bits());
      c = (powermod_g_p(k) + f) % q;
      d = (k - x * c) % q;

      if(c.is_zero())
         continue;

      SecureVector<byte> output(2*q.bytes());
      c.binary_encode(output + (q.bytes() - c.bytes()));
      d.binary_encode(output + output.size() / 2 + (q.bytes() - d.bytes()));
      return output;
      }
   }

}
